const { logger, stringify } = require('./helpers/logger')

function queryDatabase (resourceBase, query) {
  const results = []
  Object.keys(resourceBase).forEach(id => {
    const isValid = Object.keys(query).reduce(
      (proposition, key) =>
        proposition && resourceBase[id].value.metadata[key] === query[key],
      true
    )
    if (isValid) results.push(resourceBase[id])
  })
  return results
}

function filterByConstraint (resources, constraint) {
  // The `constraint` object should only be one property, and it shouln't be `value`
  if (constraint.value) {
    logger.warn(
      `There was an attempt to query [an entry's metadata] using '.getAll()'. Use '.query()' instead. Constraint - ${stringify(
        constraint
      )}`
    )
    delete constraint.value
  }

  return resources.filter(resource =>
    Object.keys(constraint).reduce(
      (proposition, key) => proposition && resource[key] === constraint[key],
      true
    )
  )
}

exports.createDatabase = () => {
  /*
    This database will be used to store resources once jet state is completely removed.
    Currently we've only moved users over to the new database, partners are still using
    the jetBroker (which used to be the dataBroker)
  */
  const users = {}
  return {
    query (resourceType, query) {
      switch (resourceType) {
        case 'users':
          return queryDatabase(users, query)
        default:
          throw new Error(
            `Failed to query resource; resource type "${resourceType}" is not supported.`
          )
      }
    },
    get (resourceType, id) {
      switch (resourceType) {
        case 'users':
          return users[id]
        default:
          throw new Error(
            `Failed to get resource; resource type "${resourceType}" is not supported.`
          )
      }
    },
    getAll (resourceType, constraint) {
      switch (resourceType) {
        case 'users':
          const allUsers = Object.values(users)
          return constraint
            ? filterByConstraint(allUsers, constraint)
            : allUsers
        default:
          throw new Error(
            `Failed to getAll resource; resource type "${resourceType}" is not supported.`
          )
      }
    },
    set (resourceType, id, data) {
      switch (resourceType) {
        case 'users':
          if (Object.keys(data).includes('id')) {
            logger.warn(
              `There was an attempt to modify the ID of a resource, which is forbidden. The ID "${
                data.id
              }" will be discarded.`
            )
          }
          users[id] = Object.assign(users[id], data, { id })
          return
        default:
          throw new Error(
            `Failed to set resource; resource type "${resourceType}" is not supported.`
          )
      }
    },
    add (resourceType, data) {
      switch (resourceType) {
        case 'users':
          if (users[data.id]) {
            logger.warn(
              `User with ID "${
                data.id
              } already exists, use .set() next time. Overwriting..."`
            )
            logger.debug(`Existing user: ${stringify(users[data.id])}`)
          }
          users[data.id] = data
          return
        default:
          throw new Error(
            `Failed to add resource; resource type "${resourceType}" is not supported.`
          )
      }
    },
    remove (resourceType, id) {
      switch (resourceType) {
        case 'users':
          logger.warn(
            'Deleting users is unsupported, this is action is being ignored.'
          )
          return
        default:
          throw new Error(
            `Failed to delete resource; resource type "${resourceType}" is not supported.`
          )
      }
    }
  }
}
