const { logger, stringify } = require('./helpers/logger')
const jet = require('node-jet')
const { createReservationManager } = require('./reservationManager')
const loadUsers = require('./helpers/loadUsers')
const path = require('path')
const fs = require('fs-extra')
const argv = require('yargs')
  .command('', 'Opens an HTTP server and accepts WS connections for node-jet.')
  .example('$0 -p 8090')
  .options({
    p: {
      alias: 'port',
      demandOption: false,
      default: '8090',
      describe: 'TCP port to use for HTTP and WS',
      type: 'number'
    },
    k: {
      alias: 'key',
      demandOption: false,
      describe:
        'The path to the key for SSL. Must provide --cert as well. Using these options will cause the server to only accept secure connections.'
    },
    c: {
      alias: 'cert',
      demandOption: false,
      describe:
        'The path to the certificate for SSL. Must provide --key as well. Using these options will cause the server to only accept secure connections.'
    },
    u: {
      alias: 'users',
      demandOption: false,
      describe: 'The path to a `.json` of user account information.'
    }
  })
  .check(argv => {
    if (argv.key || argv.cert) {
      if (!argv.key || !argv.cert) {
        throw Error(
          `ERROR: Using the key or cert option requires both to be provided!`
        )
      }
    }
    return true
  })
  .help('h')
  .alias('h', 'help').argv

process.on('unhandledRejection', (reason, promise) =>
  logger.warn(
    `Unhandled Rejection at: ${stringify(promise)}, reason: ${reason}`
  )
)

let httpServer

if (argv.key && argv.cert) {
  const https = require('https')
  const options = {
    key: fs.readFileSync(path.resolve(argv.key)),
    cert: fs.readFileSync(path.resolve(argv.cert))
  }
  // Create Webserver and start listening on given or default port
  httpServer = https.createServer(options)
} else {
  const http = require('http')
  // Create Webserver and start listening on given or default port
  httpServer = http.createServer()
}

const start = async () => {
  // Start http server
  httpServer.listen(argv.port)

  // Create Jet Daemon
  const daemon = new jet.Daemon()

  // Listen for connections
  daemon.on('connection', function (peer) {
    logger.info(`Peer Connected, ID: ${peer.id}`)
  })

  daemon.on('disconnect', function (peer) {
    logger.info(`Peer Disconnected, ID: ${peer.id}`)
  })

  try {
    // Create the local peer to handle state
    const localPeer = daemon.createLocalPeer()
    await localPeer.connect()

    // Create reservation manager
    const manager = await createReservationManager(localPeer, daemon)

    // Load users
    if (argv.users) {
      await loadUsers(manager, argv.users)
    } else {
      logger.warn("Path to users' .json file not provided")
    }

    await daemon.listen({
      server: httpServer
    })
  } catch (error) {
    logger.error(`Error on server start - ${error}`)
    process.exit(1)
  }
}

// Start server
start()
