const { logger, stringify } = require('./helpers/logger')
const jet = require('node-jet')
const { createJetBroker } = require('./jetBroker')
const { createDatabase } = require('./database')
const { createQueueManager } = require('./queueManager')
const { createRequest } = require('./helpers/request')
const {
  createPartnerManager,
  createUserManager
} = require('./resourceManagers')
const parseResourcePath = require('./helpers/parseResourcePath')

exports.createReservationManager = async (peer, daemon) => {
  const jetBroker = createJetBroker(peer)
  const database = createDatabase()

  const queueManager = createQueueManager()

  const resourceManagers = {
    partners: createPartnerManager(peer, queueManager.onResourceProvided),
    users: createUserManager(queueManager.onResourceProvided)
  }

  const methods = {
    addResource ({ type, data }) {
      logger.verbose(`Adding resource of type "${type}"`)

      if (type === 'users') {
        resourceManagers[type].add(database, data)
        // database.startTransaction(db => resourceManagers[type].add(db, data))
      } else {
        jetBroker.startTransaction(jb => resourceManagers[type].add(jb, data))
      }
    },
    releaseResource ({ type, id }) {
      logger.verbose(`Releasing resource of type "${type}"`)

      if (type === 'users') {
        resourceManagers[type].release(database, id)
        // database.startTransaction(db => resourceManagers[type].release(db, id))
      } else {
        jetBroker.startTransaction(jb => resourceManagers[type].release(jb, id))
      }
    },
    requestResources (rawRequest) {
      logger.info(`Request received from ${rawRequest.from}`)
      const request = createRequest(peer, rawRequest)
      logger.debug(`Request: ${stringify(request)}`)

      // Timeout the request after 10 minutes
      request.timeout.promise.then(async () => {
        cancelRequest(request.from)
        request.cancel()
        logger.warn(`Request from ${request.from} cancelled due to timeout`)
      })

      // Push converted request into the queue
      queueManager.push(request) // TODO: Move this back inside the transaction below once we get rid of jetBroker

      // Start transaction to prevent other actions on the queue and data
      jetBroker.startTransaction(jb => {
        // Process queue for partners
        return queueManager.onResourceProvided(jb, false)
      })
      queueManager.onResourceProvided(database, true)
    }
  }

  function cancelRequest (peerId) {
    logger.verbose(`Cancelling requests from ${peerId}`)

    // Remove any requests in the queue belonging to the peer
    const removedRequests = queueManager.removeFromQueue(peerId)

    // TODO: remove this once we remove node-jet
    jetBroker.startTransaction(async jb => {
      // Find all resources reserved by peer that disconnected
      const query = {
        valueField: {
          reservedBy: {
            equals: peerId
          }
        }
      }
      const reservedByPeer = await jb.get(query)

      if (reservedByPeer && reservedByPeer.length) {
        await Promise.all(
          reservedByPeer.map(resource => {
            // Get the type of resource and resource id from the path
            const { type, id } = parseResourcePath(resource.path)
            return resourceManagers[type].release(jb, id)
          })
        )
      }
    })

    // TODO: Do the same as below but with partners as well (once we remove node-jet)
    // Find all users reserved by the peer which disconnected
    const orphanedUsers = database.getAll('users', { reservedBy: peerId })

    // Release all those users
    orphanedUsers.forEach(user =>
      resourceManagers['users'].release(database, user.id)
    )

    logger.info(
      `Cancelled ${removedRequests.length} request(s) belonging to ${peerId}`
    )
    logger.debug(`Cancelled requests: ${stringify(removedRequests)}`)
  }

  // Create jet Methods
  const addResourceMethod = new jet.Method(
    '/werkout-server/methods/addResource'
  ).on('call', methods.addResource)

  const releaseResourceMethod = new jet.Method(
    '/werkout-server/methods/releaseResource'
  ).on('call', methods.releaseResource)

  const requestResourcesMethod = new jet.Method(
    '/werkout-server/methods/requestResources'
  ).on('call', methods.requestResources)

  // Add listeners to daemon
  daemon.on('disconnect', peer => {
    cancelRequest(peer.id)
  })

  // Add methods
  await Promise.all([
    peer.add(addResourceMethod),
    peer.add(releaseResourceMethod),
    peer.add(requestResourcesMethod)
  ])

  return methods
}
