const { logger, stringify } = require('./helpers/logger')
const generateId = require('shortid')

exports.createPartnerManager = function (peer, onResourceProvided) {
  return {
    async add (jb, resourceData) {
      await jb.set('partners', resourceData)
      logger.info(`Partner added successfully. ID: ${resourceData.id}`)
      logger.verbose(
        `Partner metadata: ${stringify(resourceData.value.metadata)}`
      )
      logger.debug(`Partner value: ${stringify(resourceData.value)}`)
      return onResourceProvided(jb, false)
    },
    async release (jb, id) {
      try {
        logger.verbose(`Releasing partner. ID: ${id}`)
        await peer.call(`/partners/${id}/methods/wo-deregister`)
        logger.info(`Partner released. ID: ${id}`)
      } catch (error) {
        logger.warn(
          `Partner has already been deregistered. ID (when it was registered): ${id}`
        )
        logger.debug(`Reported error when attempting to deregister: ${error}`)
      }
    }
  }
}

exports.createUserManager = function (onResourceProvided) {
  return {
    add (db, data) {
      const id = generateId()
      db.add('users', {
        id,
        isAvailable: true,
        value: data
      })
      logger.info(`User added successfully. ID: ${id}`)
      logger.debug(`User value: ${stringify(data)}`)
    },
    async release (db, id) {
      const user = db.get('users', id)
      logger.debug(`User: ${stringify(user)}`)

      if (!user) {
        logger.error(
          `Failed to get a user with query while trying to release it. Target ID: ${id}`
        )

        const allUsers = db.getAll('users')
        logger.debug(`All users: ${stringify(allUsers)}`)

        return onResourceProvided(db, true)
      }

      try {
        db.set('users', id, { isAvailable: true, reservedBy: null })
        logger.info(`User released successfully. ID: ${id}`)
      } catch (error) {
        logger.error(`Failed to release user: ${error}`)
      }

      return onResourceProvided(db, true)
    }
  }
}
