const fs = require('fs-extra')
const generateId = require('shortid')

const { createDatabase } = require('../lib/database')

const USERS = fs.readJSONSync('../../users.json')

function mockDatabaseEntry (resource) {
  return {
    id: generateId(),
    isAvailable: true,
    value: resource
  }
}

function loadUsers (database) {
  const entries = {}
  USERS.forEach(user => {
    const entry = mockDatabaseEntry(user)
    database.add('users', entry)
    entries[entry.id] = entry
  })
  return entries
}

// .getAll() tests
test('[.getAll()] Get all resources of a type that is not supported by the database', () => {
  const database = createDatabase()

  expect(() => {
    database.getAll('cars')
  }).toThrow()
})

// TODO: write the same test but for partners (once we get rid of node-jet)
test('[.getAll()] Get all users', () => {
  const database = createDatabase()
  const dbCopy = loadUsers(database)

  const allUsers = database.getAll('users')

  expect(allUsers).toEqual(Object.values(dbCopy))
})

// TODO: write the same test but for partners (once we get rid of node-jet)
test('[.getAll()] Get all users in empty database', () => {
  const database = createDatabase()

  expect(database.getAll('users')).toEqual([])
})

// TODO: write the same test but for partners (once we get rid of node-jet)
test('[.getAll()] Get all users that fit a constraint', () => {
  const database = createDatabase()
  loadUsers(database)
  const JEST_ID = 'jestTest'

  const matchedUsers = database.query('users', { colour: 'blue' })
  database.set('users', matchedUsers[2].id, { reservedBy: JEST_ID })
  database.set('users', matchedUsers[4].id, { reservedBy: JEST_ID })
  database.set('users', matchedUsers[6].id, { reservedBy: JEST_ID })

  const reservedUsers = [
    { ...matchedUsers[2], reservedBy: JEST_ID },
    { ...matchedUsers[4], reservedBy: JEST_ID },
    { ...matchedUsers[6], reservedBy: JEST_ID }
  ]

  expect(database.getAll('users', { reservedBy: JEST_ID })).toEqual(
    reservedUsers
  )
})

// .get() tests
test('[.get()] Get a type of resource that does not exist', () => {
  const database = createDatabase()

  expect(() => {
    database.get('fruits', 'apple')
  }).toThrow()
})

// TODO: write the same test but for partners (once we get rid of node-jet)
test('[.get()] Get a user', () => {
  const database = createDatabase()

  const user = mockDatabaseEntry(USERS[0])
  database.add('users', user)

  expect(database.get('users', user.id)).toEqual(user)
})

// TODO: write the same test but for partners (once we get rid of node-jet)
test('[.get()] Get a user that does not exist', () => {
  const database = createDatabase()

  let user = database.get('users', 'randomID')
  expect(user).toBe(undefined)
})

// .add tests
test('[.add()] Add a resource whose type is not supported by the database', () => {
  const database = createDatabase()

  const fakeResource = {
    id: '22',
    value: {
      color: 'red'
    }
  }

  expect(() => {
    database.add('colours', fakeResource)
  }).toThrow()
})

// TODO: write the same test but for partners (once we get rid of node-jet)
test('[.add()] Add a user to the database', () => {
  const database = createDatabase()

  const user = mockDatabaseEntry(USERS[1])
  database.add('users', user)

  let allUsers = database.getAll('users')

  expect(allUsers.length).toBe(1)
  expect(allUsers[0]).toEqual(user)
})

// .set() tests
test('[.set()] Set a resource whose type is not support by the database', () => {
  const database = createDatabase()

  expect(() => {
    database.set('computers', '#FA12B2', { RAM: '12GB' })
  }).toThrow()
})

// TODO: write the same test but for partners (once we get rid of node-jet)
test("[.set()] Change a user's state", () => {
  const database = createDatabase()

  const user = mockDatabaseEntry(USERS[2])
  database.add('users', user)

  database.set('users', user.id, { isAvailable: false })

  expect(database.get('users', user.id)).toEqual({
    ...user,
    isAvailable: false
  })
})

// TODO: write the same test but for partners (once we get rid of node-jet)
test('[.set()] Set a user that does not exist in the database', () => {
  const database = createDatabase()

  expect(() => {
    database.set('users', '#abcdefg', { isAvailable: false })
  }).toThrow()
})

// .remove() tests
test('[.remove()] Remove a resource whose type is not supported by the database', () => {
  const database = createDatabase()

  expect(() => {
    database.remove('fluids', 'H20')
  }).toThrow()
})

// TODO: write a test for removing partners

// .query() tests
test('[.query()] Query a resource whose type is not supported by the database', () => {
  const database = createDatabase()

  expect(() => {
    database.query('plants', { color: 'green' })
  }).toThrow()
})

// TODO: write the same test but for partners (once we get rid of node-jet)
test('[.query()] Query for a user', () => {
  const database = createDatabase()
  const copyOfDatabase = loadUsers(database)

  const matchedUsers = database.query('users', { colour: 'blue' })

  const blueUsers = Object.values(copyOfDatabase).filter(
    ({ value }) => value.metadata.colour === 'blue'
  )

  expect(matchedUsers).toEqual(blueUsers)
})

// TODO: write the same test but for partners (once we get rid of node-jet)
test('[.query()] Query for a user that does not exist', () => {
  const database = createDatabase()

  const matchedUsers = database.query('users', { colour: 'yellow' })

  expect(matchedUsers).toEqual([])
})
